﻿using Npgsql;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace TapfloDB.View.TestRules
{
    public class TestRule
    {
        #region Properties

        public int ID { get; private set; }

        private string pattern;
        private Regex rule;
        public String Rule {
            get { return pattern; }
            set { rule = new Regex(value); pattern = value; }
        }

        public String Description { get; private set; }
        public int?[] Required { get; private set; }
        public bool General { get; private set; }

        #endregion

        #region Static

        public static TestRule Default { get; private set; }

        private static DateTime lastUpdate;

        private static Dictionary<int, TestRule> builtin = new Dictionary<int, TestRule>();
        public static TestRule[] BuiltIn {
            get { return builtin.Values.ToArray(); }
        }

        private static Dictionary<int, TestRule> rules = new Dictionary<int, TestRule>();
        public static TestRule[] AllRules {
            get { return rules.Values.Union(builtin.Values).ToArray(); }
        }

        static TestRule(){
            Default = new TestRule(1, "", "Domyślna Reguła Testów (nieobowiązkowe testy: \"Tłumienie\" i \"Wandersa\")", new int?[] {0, -8, -9});
            lastUpdate = DateTime.Today.Subtract(TimeSpan.FromDays(1));

            Update(true);
        }

        public static bool Update(bool force) {
            if (!force && DateTime.Now.Subtract(lastUpdate).TotalMinutes < 30)
                return false;

            Server.Connect();
            NpgsqlDataReader dr = null;
            TestRule tr;
            try {
                NpgsqlCommand cmd = Server.Command("SELECT id, pattern, description, required::text FROM tdb.\"TestRules\" ORDER BY id;");
                dr = cmd.ExecuteReader();
                if (!dr.HasRows) return false;

                builtin.Clear();
                rules.Clear();
                lastUpdate = DateTime.Now;

                while (dr.Read()){
                    tr = new TestRule((int)dr["id"], (string)dr["pattern"], (string)dr["description"], PSqlArrays.ParseInt(dr["required"], 1));
                    if (tr.ID == 1)
                        Default = tr;
                    else if (tr.ID < 100)
                        builtin.Add(tr.ID, tr);
                    else
                        rules.Add(tr.ID, tr);
                }
            }
            catch (Exception exp){
                Server.Log("Błąd inicjalizacji listy reguł Testów - {0}", exp.Message);
                throw exp;
            }
            finally {
                if(dr != null)
                    dr.Close();
            }

            return true;
        }

        public static TestRule Find(string product) {
            product = product.ToUpper();
            foreach (TestRule tr in TestRule.AllRules){
                if (tr.rule.IsMatch(product))
                    return tr;
            }

            return Default;
        }

        public static TestRule Add(int id, string pattern, string description, int?[] required){
            TestRule tr = new TestRule(id, pattern, description, required);
            if (rules.ContainsKey(id))
                rules.Remove(id); // Remove old instance
            rules.Add(id, tr);
            return tr;
        }

        #endregion

        public TestRule(int id, string pattern, string description, int?[] required) {
            this.ID = id;
            this.Rule = pattern;
            this.Description = description;
            this.Required = required;
            // General requirement
            var gen = from r in Required where !r.HasValue select r;
            this.General = gen.Count() == 0;
        }

        // 0 - everythind is required
        // null - nothing is required
        // +value - abs(value) == testId is required
        // -value - abs(value) == testId is NOT required
        public bool IsRequired(int testId) {
            // Use specific requirement in rule
            int? spec = (from id in Required where id.HasValue && Math.Abs(id.Value) == testId
                         orderby id.Value select id).FirstOrDefault();
            if (spec.HasValue)
                return spec.Value > 0;

            // No specific use general rule
            return General;
        }

        public override string ToString(){
            return pattern;
        }
    }
}
